//#define _GLIBCXX_PARALLEL

// Usage: ./binloader IN1 IN2 IN3 ... OUT NUM_BUCKETS NUM_POS_PER_SUBSHARD

#include <stdio.h>
#include <vector>
#include <mtbl.h>
#include <algorithm>
#include <utility>
#include <memory>
#include <string>
#include <unordered_set>
#include <string.h>
#include "count.pb.h"
#include "arena.h"
#include "hash.h"

#define DUMMY_TIMESTAMP 32503680000

using namespace std;

Arena arena;

enum Result { WHITE = 0, DRAW, BLACK };
struct Element {
	char *bpfen;
	int bpfen_len;
	char move[8];   // Na1xc3+
	Result result;
	int opening_num, white_elo, black_elo;
	int file_num;
	time_t timestamp;
	long start_position;

	bool operator< (const Element& other) const {
		int shared_len = min(bpfen_len, other.bpfen_len);
		int s = memcmp(bpfen, other.bpfen, shared_len);
		if (s < 0) {
			return true;
		} else if (s > 0) {
			return false;
		} else {
			return bpfen_len < other.bpfen_len;
		}
	}
};

struct ShardData {
	vector<Element> elems;
	unique_ptr<Arena> arena;  // Used to allocate bpfen.
	int num_written_subshards = 0;
};

void write_subshard(const char *basename, ShardData* shard, int bucket)
{
	string buf;  // Keep allocated.
	char filename[256];
	snprintf(filename, sizeof(filename), "%s.part%04d.subshard%04d",
		basename, bucket, shard->num_written_subshards++);
	printf("Writing SSTable %s...\n", filename);

	sort(shard->elems.begin(), shard->elems.end());

	mtbl_writer_options* wopt = mtbl_writer_options_init();
	mtbl_writer_options_set_compression(wopt, MTBL_COMPRESSION_SNAPPY);
	mtbl_writer* mtbl = mtbl_writer_init(filename, wopt);
	Count c;
	unordered_set<string> moves;
	for (size_t i = 0; i < shard->elems.size(); ++i) {
		const Element &e = shard->elems[i];
		if (e.result == WHITE) {
			c.set_white(c.white() + 1);
		} else if (e.result == DRAW) {
			c.set_draw(c.draw() + 1);
		} else if (e.result == BLACK) {
			c.set_black(c.black() + 1);
		}
		if (e.white_elo >= 100 && e.black_elo >= 100) {
			c.set_sum_white_elo(c.sum_white_elo() + e.white_elo);
			c.set_sum_black_elo(c.sum_black_elo() + e.black_elo);
			c.set_num_elo(c.num_elo() + 1);
		}
		if (!c.has_first_timestamp() || e.timestamp < c.first_timestamp()) {
			if (e.timestamp != DUMMY_TIMESTAMP) {
				c.set_first_timestamp(e.timestamp);
			}
			c.set_opening_num(e.opening_num);
			c.set_pgn_file_num(e.file_num);
			c.set_pgn_start_position(e.start_position);
		}
		if (!moves.count(e.move)) {
			moves.insert(e.move);
			c.add_move(e.move);
		}
		if (i == shard->elems.size() - 1 ||
		    e.bpfen_len != shard->elems[i + 1].bpfen_len ||
		    memcmp(e.bpfen, shard->elems[i + 1].bpfen, e.bpfen_len) != 0) {
			c.SerializeToString(&buf);
			mtbl_writer_add(mtbl,
				(const uint8_t *)e.bpfen, e.bpfen_len,
				(const uint8_t *)buf.data(), buf.size());
			c = Count();
			moves.clear();
		}
	}
	mtbl_writer_destroy(&mtbl);

	shard->elems.clear();
	shard->arena.reset(new Arena);
}

int main(int argc, char **argv)
{
	int num_buckets = atoi(argv[argc - 2]);
	size_t num_pos_per_subshard = atoi(argv[argc - 1]);  // 500000 is a reasonable value.

	vector<ShardData> shards;
	shards.resize(num_buckets);

	for (int i = 0; i < num_buckets; ++i) {
		shards[i].elems.reserve(num_pos_per_subshard);
		shards[i].arena.reset(new Arena);
	}

	size_t num_elems = 0;
	for (int i = 1; i < argc - 3; ++i) {
		FILE *fp;
		if (strcmp(argv[i], "-") == 0) {
			fp = stdin;
		} else {
			fp = fopen(argv[i], "rb");
			if (fp == NULL) {
				perror(argv[i]);
				exit(1);
			}
		}
		for ( ;; ) {
			char bpfen[256];

			int bpfen_len = getc(fp);
			if (bpfen_len == -1) {
				break;
			}
			if (bpfen_len >= int(sizeof(bpfen))) {
				fprintf(stderr, "Overlong BPFEN (%d bytes)\n", bpfen_len);
		//		exit(1);
				break;
			}
			if (fread(bpfen, bpfen_len, 1, fp) != 1) {
				perror("fread()");
		//		exit(1);
				break;
			}

			int r = getc(fp);
			if (r == -1) {
				perror("getc()");
				//exit(1);
				break;
			}

			int opening_num, white_elo, black_elo, file_num;
			time_t timestamp;
			long start_position;
			if (fread(&white_elo, sizeof(white_elo), 1, fp) != 1) {
				perror("fread()");
				//exit(1);
				break;
			}
			if (fread(&black_elo, sizeof(black_elo), 1, fp) != 1) {
				perror("fread()");
				//exit(1);
				break;
			}
			if (fread(&opening_num, sizeof(opening_num), 1, fp) != 1) {
				perror("fread()");
				//exit(1);
				break;
			}
			if (fread(&timestamp, sizeof(timestamp), 1, fp) != 1) {
				perror("fread()");
				//exit(1);
				break;
			}
			if (fread(&file_num, sizeof(file_num), 1, fp) != 1) {
				perror("fread()");
				//exit(1);
				break;
			}
			if (fread(&start_position, sizeof(start_position), 1, fp) != 1) {
				perror("fread()");
				//exit(1);
				break;
			}


			char move[8];
			int l = getc(fp);
			if (l == -1) {
				break;
			}
			if (l >= int(sizeof(move))) {
				fprintf(stderr, "Overlong move (%d bytes)\n", l);
		//		exit(1);
				break;
			}
			if (fread(&move[0], l, 1, fp) != 1) {
				perror("fread()");
		//		exit(1);
				break;
			}
			move[l] = 0;

			int bucket = hash_key_to_bucket(bpfen, bpfen_len, num_buckets);
			Element e;
			e.bpfen = shards[bucket].arena->alloc(bpfen_len);
			memcpy(e.bpfen, bpfen, bpfen_len);
			e.bpfen_len = bpfen_len;
			strcpy(e.move, move);
			e.result = Result(r);
			e.opening_num = opening_num;
			e.white_elo = white_elo;
			e.black_elo = black_elo;
			e.file_num = file_num;
			e.timestamp = timestamp;
			e.start_position = start_position;
			shards[bucket].elems.push_back(e);
			++num_elems;

			if (shards[bucket].elems.size() >= num_pos_per_subshard) {
				write_subshard(argv[argc - 3], &shards[bucket], bucket);
				shards[bucket].elems.reserve(num_pos_per_subshard);
			}
		}
		fclose(fp);

		printf("Read %ld elems\n", num_elems);
	}

	for (int i = 0; i < num_buckets; ++i) {
		write_subshard(argv[argc - 3], &shards[i], i);
	}
}

